@@ -25,6 +25,7 @@ import shutil
 import gzip
 import inspect
 import glob
+import traceback
 
 import threading
 
@@ -83,6 +84,27 @@ def json_load( fn ) :
 def json_loads( str ) :
     return( json.loads( str ) )
 
+def json_stdin( ) :
+
+    data_in_list = sys.stdin.readlines( )
+    data_in_json = "".join( data_in_list )
+
+    data_in = json_loads(data_in_json)
+
+    return( data_in )
+
+def json_dumps( inputs ) :
+    return( json.dumps( inputs ) )
+
+#def json_pretty( json_in ) :
+#    data = json.loads(json_in)
+#    pretty_object = json.dumps(data, indent=4)
+#    print(pretty_object)
+
+def json_prettyprint( data ) :
+    pretty_object = json.dumps(data, indent=4)
+    print(pretty_object)
+
 #####################################################################
 
 # FIXME TODO aaarrggghhh!!!
@@ -124,6 +146,9 @@ def pickle_save( datakey , data , mvtotmp = False ) :
     if( mvtotmp ) :
         os.rename( fn , "/tmp/" + fn )
 
+def get_cwd( ) :
+    directory_path = os.getcwd()
+    return(directory_path)
 
  
 def pathnotallowed( p ) :
@@ -195,11 +220,25 @@ def printexit( msg ) :
 
 
 def printexception( e = False ) :
-    if e:
+    if e :
         print( "EXCEPTION" + str( e ) )
     etype , value, tb = sys.exc_info( )
     info , error = format_exception( etype , value , tb )[ -2: ]
     print( f'Exception in:\n{info}\n{error}' )
+    print( full_stack( ) )
+
+def full_stack():
+    exc = sys.exc_info()[0]
+    if exc is not None:
+        f = sys.exc_info()[-1].tb_frame.f_back
+        stack = traceback.extract_stack(f)
+    else:
+        stack = traceback.extract_stack()[:-1]  # last one would be full_stack()
+    trc = 'Traceback (most recent call last):\n'
+    stackstr = trc + ''.join(traceback.format_list(stack))
+    if exc is not None:
+        stackstr += '  ' + traceback.format_exc().lstrip(trc)
+    return stackstr
 
 
 heartbeat_last = 0
@@ -433,22 +472,52 @@ def datetime_now_db( td = 0 ) :
 
     d = datetime.now( ) + timedelta( seconds = td)
 
-    return(d.strftime('%Y-%m-%d %H:%M:%S'))
+    return( d.strftime('%Y-%m-%dT%H:%M:%S'))
 
 def datetime_timedelta_db( dbdtstr , td ) :
-    d = datetime.strptime(dbdtstr, '%Y-%m-%d %H:%M:%S') + timedelta(seconds=td)
-    return(d.strftime('%Y-%m-%d %H:%M:%S'))
+    d = datetime.strptime( dbdtstr , "%Y-%m-%dT%H:%M:%S" ) + timedelta( seconds = td )
+    return( d.strftime( "%Y-%m-%dT%H:%M:%S" ) )
 
 def datetime_inrange_db( dbdt1 , dbdt2 , td) :
-    d1 = datetime.strptime(dbdt1, '%Y-%m-%d %H:%M:%S')  
-    d2 = datetime.strptime(dbdt2, '%Y-%m-%d %H:%M:%S') 
+    d1 = datetime.strptime(dbdt1, '%Y-%m-%dT%H:%M:%S')  
+    d2 = datetime.strptime(dbdt2, '%Y-%m-%dT%H:%M:%S') 
 
     dtdiff=d2-d1
     dtdiff_seconds=int(dtdiff.total_seconds())
     return( dtdiff_seconds <= td )
 
+
+def datetime_centerwindow( p , ec , ew ):
+    a = datetime_timedelta_db( p[ ec ] , -0.5 * p[ ew ] )
+    b = datetime_timedelta_db( p[ ec ] , 0.5 * p[ ew ] )
+    return([a,b])
+
+
+def datetime_initstartend( p , es , ee , ew ) :
+    if( ( not es in p) and ( not ee in p ) ) :
+        a = datetime_now_db( -1 * p[ ew ] )
+        b = datetime_now_db( )
+        return([a,b])
+    elif( not ee in p) :
+        b = datetime_timedelta_db( p[ es ] , p[ ew ] )
+        return([p[es],b])
+    elif( not es in p) :
+        a = datetime_timedelta_db( p[ ee ] , -1 * p[ ew ] )
+        return([a,p[ee]])
+        
+    return([p[es],p[ee]])
+ 
+
+
 ################
 
+def dir_exists( dp ) :
+    if( not os.path.isdir( dp ) ) :
+        return( False )
+    return( True )
+
+def dir_make( dp ) :
+    os.makedirs(dp, exist_ok=True)
 # Return file getcontents given filename and parent script __FILE__
 def file_getcontent( fn , scp ) :
 
@@ -461,6 +530,8 @@ def file_getcontent( fn , scp ) :
 
     return( contents )
 
+
+
 def file_exists( fp ) :
     return( os.path.isfile(fp) )
 
@@ -495,3 +566,19 @@ def thread_get_id( ) :
 def glob_files( dp ) :
     return( glob.glob( dp ) )
 
+
+def match_listindict( l , d ) :
+    res = [ ]
+    for dk in l :
+        if dk in d :
+            res.append( dk )
+
+    return( list( dict.fromkeys( res ) ) )   
+
+
+def math_dictkeylimit( d , k , l ) :
+    if( not k in d ) :
+        return( l )
+    elif ( int( d[ k ] ) > l ) :
+        return( l )
+    return(d[k])
